/**************************************************************************
ColourizeSmooth uses an approximation to a  general colourizing algorithm
to smooth a given clip.  Algorithm information can be found at
http://www.cs.huji.ac.il/~yweiss/Colorization.
Copyright (C) 2005 Aditya Bhargava

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************/

#include "ColourizeSmooth.h"

PVideoFrame __stdcall ColourizeSmooth::GetFrame(int n, IScriptEnvironment* env)
{	
	PVideoFrame output = env->NewVideoFrame(vi);

	int largestChange;
	
	YUYVToYUV(child->GetFrame(n, env));
	
	if (iterations)
		for (int i = 0; i < iterations; i++)
		{
			largestChange = doIteration();
			
			copyBuffer(env);
		}
	else
		do
		{
			largestChange = doIteration();
			
			copyBuffer(env);
		}while(largestChange > threshold);

	YUVToYUYV(output);
	
	return output;
}

bool ColourizeSmooth::areNeighbours(int x1, int y1, int x2, int y2)
{
	int distSq = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);

	if ((distSq != 0) && (distSq <= (radius * radius)))
		return true;
	else
		return false;
}

void ColourizeSmooth::copyBuffer(IScriptEnvironment* env)
{
	const unsigned char* pSrc = pBuffer2;
	unsigned char* pDst = pBuffer1;
	int temp = 3 * width * height;

	env->BitBlt(pDst, temp, pSrc, temp, temp, 1);
}

int ColourizeSmooth::doIteration()
{
	int largestChange = INT_MIN;
	unsigned char* it1 = pBuffer1;
	unsigned char* it2 = pBuffer2;
	
	for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
		for (int x = 0; x < width; x++)
		{
			int sum = 0, squareSum = 0, neighbours = 0, change = 0;
			double mean = 0, varSq = 0, weight = 0, weightSum = 0, weightedAverageU = 0, weightedAverageV = 0;
			unsigned char *it3, newU, newV;
			
			for (int wy = y - radius; wy <= y + radius; wy++)
			{
				int useWY = abs(height - 1 - abs(height - 1 - wy));

				it3 = pBuffer1 + useWY * 3 * width;

				for (int wx = x - radius; wx <= x + radius; wx++)
				{
					int useWX = abs(width - 1 - abs(width - 1 - wx));

					if (areNeighbours(x, y, useWX, useWY))
					{
						sum += it3[3 * useWX];
						squareSum += (it3[3 * useWX] * it3[3 * useWX]);
						
						neighbours++;
					}
				}
			}

			mean = static_cast<double>(sum) / static_cast<double>(neighbours);
			varSq = static_cast<double>(squareSum) / static_cast<double>(neighbours) - mean * mean;

			for (int wy = y - radius; wy <= y + radius; wy++)
			{
				int useWY = abs(height - 1 - abs(height - 1 - wy));

				it3 = pBuffer1 + useWY * 3 * width;

				for (int wx = x - radius; wx <= x + radius; wx++)
				{
					int useWX = abs(width - 1 - abs(width - 1 - wx));

					if (areNeighbours(x, y, useWX, useWY))
					{
						weight = weighter(it1, x, it3, useWX, mean, varSq);
						weightSum += weight;

						weightedAverageU += (weight * it3[3 * useWX + 1]);
						weightedAverageV += (weight * it3[3 * useWX + 2]);
					}
				}
			}

			newU = static_cast<unsigned char>(weightedAverageU / weightSum);
			newV = static_cast<unsigned char>(weightedAverageV / weightSum);

			change = abs(it1[3 * x + 1] - newU);
			if (change > largestChange)
				largestChange = abs(it1[3 * x + 1] - newU);

			change = abs(it1[3 * x + 1] - newV);
			if (change > largestChange)
				largestChange = abs(it1[3 * x + 2] - newV);

			it2[3 * x] = it1[3 * x];
			it2[3 * x + 1] = newU;
			it2[3 * x + 2] = newV;
		}

	return largestChange;
}

double ColourizeSmooth::weighter(unsigned char* p1, int x1, unsigned char* p2, int x2, double mean, double varSq)
{
	if (varSq == 0.0)
		return 1.0;
	if (gaussian)
		return pow(M_E, -1 * (p1[3 * x1] - p2[3 * x2]) * (p1[3 * x1] - p2[3 * x2]) / (2 * varSq));
	else
		return (1 + (p1[3 * x1] - mean) * (p2[3 * x2] - mean) / (varSq));
}

void ColourizeSmooth::YUVToYUYV(PVideoFrame& dst)
{
	unsigned char* pDst = dst->GetWritePtr();
	unsigned char* it = pBuffer1;

	for (int y = 0; y < height; y++, pDst += dst->GetPitch(), it += 3 * width)
		for (int x = 0, x1 = 0; x < 3 * width, x1 < dst->GetRowSize(); x += 6, x1 += 4)
		{
			pDst[x1] = it[x];
			pDst[x1 + 1] = (it[x + 1] + it[x + 4] + 1) >> 1;
			pDst[x1 + 2] = it[x + 3];
			pDst[x1 + 3] = (it[x + 2] + it[x + 5] + 1) >> 1;
		}
}

void ColourizeSmooth::YUYVToYUV(PVideoFrame& src)
{
	const unsigned char* pSrc = src->GetReadPtr();
	unsigned char* it = pBuffer1;

	if (sup)
		for (int y = 0; y < height; y++, pSrc += src->GetPitch(), it += 3 * width)
			for (int x = 0, x1 = 0; x < 3 * width, x1 < src->GetRowSize(); x += 6, x1 += 4)
			{
				it[x] = pSrc[x1];
				it[x + 1] = it[x + 4] = pSrc[x1 + 1];
				it[x + 2] = it[x + 5] = pSrc[x1 + 3];
				it[x + 3] = pSrc[x1 + 2];
			}
	else
		for (int y = 0; y < height; y++, pSrc += src->GetPitch(), it += 3 * width)
		{
			int x = 0, x1 = 0;

			while ((x < 3 * (width - 2)) && (x1 < src->GetRowSize() - 4))
			{
				it[x] = pSrc[x1];
				it[x + 1] = pSrc[x1 + 1];
				it[x + 2] = pSrc[x1 + 3];

				it[x + 3] = pSrc[x1 + 2];
				it[x + 4] = (pSrc[x1 + 1] + pSrc[x1 + 5]) >> 1;
				it[x + 5] = (pSrc[x1 + 3] + pSrc[x1 + 7]) >> 1;

				x += 6;
				x1 += 4;
			}

			it[x] = pSrc[x1];
			it[x + 1] = it[x + 4] = pSrc[x1 + 1];
			it[x + 2] = it[x + 5] = pSrc[x1 + 3];
			it[x + 3] = pSrc[x1 + 2];
		}
}

AVSValue __cdecl createColourizeSmooth(AVSValue args, void* data, IScriptEnvironment* env)
{
	bool _sup = false, _gaussian = false, isYV12 = args[0].AsClip()->GetVideoInfo().IsYV12();
	int _radius = 1, _iterations = 0, _threshold = 165;
	PClip temp;

	if (isYV12)
	{
		try
		{
			temp = env->Invoke("ConvertToYUY2", args[0].AsClip()).AsClip();
		}
		catch (IScriptEnvironment::NotFound)
		{
			env->ThrowError("ColourizeSmooth: ConvertToYUY2 function not found.");
		}
	}
	else
		temp = args[0].AsClip();

	temp = new ColourizeSmooth(temp, args[1].AsBool(_sup), args[2].AsInt(_radius), args[3].AsBool(_gaussian),
		args[4].AsInt(_iterations), args[5].AsInt(_threshold), env);

	if (isYV12)
	{
		try
		{
			temp = env->Invoke("ConvertToYV12", temp).AsClip();
		}
		catch (IScriptEnvironment::NotFound)
		{
			env->ThrowError("ColourizeSmooth: ConvertToYV12 function not found.");
		}
	}

	return temp;
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("ColourizeSmooth", "c[sup]b[radius]i[gaussian]b[iterations]i[threshold]i",
		createColourizeSmooth, 0);

	return "ColourizeSmooth v0.7.4.1";
}